/* $Id: measures.c,v 1.12 2008/03/30 23:01:53 pnykanen Exp $ */

#include "measures.h"

/* for getting Euclidean distance of 2 frames */
double getEuclideanDistance(double *a, double *b, int *dim)
{
  unsigned short int i;
  double diff = 0.0;
  double squaresum = 0.0;

  for (i=0; i < *dim; i++)
    {
      diff = a[i] - b[i];
      squaresum += diff*diff;
    }
  return sqrt(squaresum);
}

/* for getting square distance of 2 frames */
double getSquareDistance(double *a, double *b, int *dim)
{
  unsigned short int i;
  double diff = 0.0;
  double squaresum = 0.0;

  for (i=0; i < *dim; i++)
    {
      diff = a[i] - b[i];
      squaresum += diff*diff;
    }
  return squaresum;
}

double point2LineSegmentDistance2D(double *p, double *line1, double *line2, int *dim)
{
  double A,B,C,D,dot,len_sq,param;
  double linepoint[2];

  A = p[0] - line1[0];
  B = p[1] - line1[1];
  C = line2[0] - line1[0];
  D = line2[1] - line1[1];

  dot = A * C + B * D;
  len_sq = C * C + D * D;
  param = dot / len_sq;

  if(param < 0)
    {
      linepoint[0] = line1[0];
      linepoint[1] = line1[1];
    }
  else if(param > 1)
    {
      linepoint[0] = line2[0];
      linepoint[1] = line2[1];
    }
  else
    {
      linepoint[0] = line1[0] + param * C;
      linepoint[1] = line1[1] + param * D;
    }

  return getEuclideanDistance(p,linepoint,dim);
}

/* for calculating distance between 2 phonemes */
double DTW(phoneme *ph1, phoneme *ph2)
{
  unsigned short int i, j;
  double dtw_distance;
  double **DTW_M;
  double *fr1, *fr2;

  DTW_M = (double**)malloc((1 + ph1->frameAmount)*sizeof(double*));
  for(i=0; i <= ph1->frameAmount; i++)
    DTW_M[i] = (double*)malloc((1 + ph2->frameAmount)*sizeof(double));

  DTW_M[0][0] = 0.0;
  for (i=1; i <= ph1->frameAmount; i++)
    DTW_M[i][0] = DBL_MAX;
  for (i=1; i <= ph2->frameAmount; i++)
    DTW_M[0][i] = DBL_MAX;

  for(i=1, fr1 = ph1->frames[0]; i <= ph1->frameAmount; fr1 = ph1->frames[i++])
    for(j=1, fr2 = ph2->frames[0]; j <= ph2->frameAmount; fr2 = ph2->frames[j++])
      DTW_M[i][j] = getEuclideanDistance(fr1,fr2,&(ph1->dimensions)) + MINOF(DTW_M[i-1][j  ],
									     DTW_M[i  ][j-1],
									     DTW_M[i-1][j-1]);
  dtw_distance = DTW_M[ph1->frameAmount][ph2->frameAmount];

  for(i=0; i <= ph1->frameAmount; i++)
    free(DTW_M[i]);
  free(DTW_M);

  return dtw_distance;
}

double closestDist(phoneme *ph1, phoneme *ph2)
{
  int i, j, k, idx;
  double distance, minimum, tmp;
  double *fr1, *fr2;

  idx = 0;
  i=0;
  fr1 = ph1->frames[0];
  fr2 = ph2->frames[0];
  distance = 0.0;
  while (i < ph1->frameAmount)
    {
      j=idx;
      k = 0;
      minimum = DBL_MAX;
      fr2 = ph2->frames[j];

      while ( k < 25 )
	{
	  tmp = getEuclideanDistance(fr1,fr2,&(ph1->dimensions));
	  if (minimum > tmp)
	    {
	      idx = j;
	      minimum = tmp;
	    }
	  j++;
	  if (j == ph2->frameAmount)
	    break;
	  fr2 = ph2->frames[j];
	  k++;
	}
  
      distance += minimum;
      i++;
      if (i == ph1->frameAmount)
	break;
      fr1 = ph1->frames[i];
    }

  distance = (distance/(double)(ph1->frameAmount));

  return distance;
}

double minPrev(double *M, int end)
{
  int i;
  double min;

  min = DBL_MAX;
  for (i=0; i <= end; i++)
    {
      if (min > M[i])
	min = M[i];
    }
 
  return min;
}

double minIdx(double *M, int end, int *idx)
{
  int i;
  double min;

  min = DBL_MAX;
  for (i=0; i <= end; i++)
    {
      if (min > M[i]) {
	min = M[i];
        *idx = i;
      }
    }
 
  return min;
}

double optimalSegmentationDistance(phoneme *ph1, phoneme *ph2)
{
  int i, j;
  double distance;
  double **M;
  double *fr1, *fr2;
  int idx;

  M = (double**)malloc((1 + ph1->frameAmount)*sizeof(double*));
  for(i=0; i <= ph1->frameAmount; i++)
    M[i] = (double*)malloc((1 + ph2->frameAmount)*sizeof(double));

  for (i=0; i <= ph1->frameAmount; i++)
    for (j=0; j <= ph2->frameAmount; j++)
      M[i][j] = DBL_MAX;

  M[0][0] = 0.0;
  
  for(i=1, fr1 = ph1->frames[0]; i <= ph1->frameAmount; fr1 = ph1->frames[i++])
    for(j=1, fr2 = ph2->frames[0]; j <= ph2->frameAmount; fr2 = ph2->frames[j++])
      M[i][j] = getEuclideanDistance(fr1,fr2,&(ph1->dimensions)) + minPrev(M[i-1],j);

  distance = M[ph1->frameAmount][ph2->frameAmount];

  /*
  idx = ph2->frameAmount;
  printf("%.2f\t",distance);
  for (i = ph1->frameAmount; i >= 1; i--){
    printf("%.2f\t",minIdx(M[i-1],idx,&idx));
  }
  */
 // getPoints(ph1, M, ph1->frameAmount, ph2->frameAmount);

  for(i=0; i <= ph1->frameAmount; i++)
    free(M[i]);
  free(M);

  distance = (distance/(double)(ph1->frameAmount));

  return distance;
}

phoneme **initialPointReduction(phoneme **phs, int n, double tolerance)
{
  int i,j,m;
  int selected[n];
  double R[n];
  
  phoneme **newPoints;

  selected[0] = 0;

  for (j=0; j < n; j++)
    {
      R[j] = DBL_MAX;
      for (i=j-1; i >= 0; i--)
	{
	  if (getEuclideanDistance(phs[0]->frames[i],phs[0]->frames[j],&(phs[0]->dimensions)) > 2*tolerance)
	    break;
	  if ( (getEuclideanDistance(phs[0]->frames[i],phs[0]->frames[j],&(phs[0]->dimensions)) < tolerance) &&
	       (R[i] + 1.0 < R[j]) )
	    {
	      R[j] = R[i] + 1.0;
	      selected[j] = i;
	    }
	}
    }

  newPoints = newPhonemes(m);

  //copy selected -> newPoints;

  return newPoints;
}

/* for calculating distance between 2 phonemes */
double DTWpath(phoneme *ph1, phoneme *ph2, double *curr_best)
{
  unsigned short int i, j;
  double dtw_distance;
  double **DTW_M;
  double *fr1, *fr2;

  DTW_M = (double**)malloc((1 + ph1->frameAmount)*sizeof(double*));
  for(i=0; i <= ph1->frameAmount; i++)
    DTW_M[i] = (double*)malloc((1 + ph2->frameAmount)*sizeof(double));

  DTW_M[0][0] = 0.0;
  for (i=1; i <= ph1->frameAmount; i++)
    DTW_M[i][0] = DBL_MAX;
  for (i=1; i <= ph2->frameAmount; i++)
    DTW_M[0][i] = DBL_MAX;

  for(i=1, fr1 = ph1->frames[0]; i <= ph1->frameAmount; fr1 = ph1->frames[i++])
    for(j=1, fr2 = ph2->frames[0]; j <= ph2->frameAmount; fr2 = ph2->frames[j++])
      DTW_M[i][j] = getEuclideanDistance(fr1,fr2,&(ph1->dimensions)) + MINOF(DTW_M[i-1][j  ],
									     DTW_M[i  ][j-1],
									     DTW_M[i-1][j-1]);
  dtw_distance = DTW_M[ph1->frameAmount][ph2->frameAmount];

  if (dtw_distance < *curr_best)
    {
      *curr_best = dtw_distance;
      backtrackPath(ph1, DTW_M, ph1->frameAmount, ph2->frameAmount);
    }

  for(i=0; i <= ph1->frameAmount; i++)
    free(DTW_M[i]);
  free(DTW_M);

  return dtw_distance;
}

double DTWpath2mat(phoneme *ph1, phoneme *ph2, char *m_cell)
{
  unsigned short int i, j;
  double dtw_distance;
  double **DTW_M;
  double *fr1, *fr2;

  DTW_M = (double**)malloc((1 + ph1->frameAmount)*sizeof(double*));
  for(i=0; i <= ph1->frameAmount; i++)
    DTW_M[i] = (double*)malloc((1 + ph2->frameAmount)*sizeof(double));

  DTW_M[0][0] = 0.0;
  for (i=1; i <= ph1->frameAmount; i++)
    DTW_M[i][0] = DBL_MAX;
  for (i=1; i <= ph2->frameAmount; i++)
    DTW_M[0][i] = DBL_MAX;

  for(i=1, fr1 = ph1->frames[0]; i <= ph1->frameAmount; fr1 = ph1->frames[i++])
    for(j=1, fr2 = ph2->frames[0]; j <= ph2->frameAmount; fr2 = ph2->frames[j++])
      DTW_M[i][j] = getEuclideanDistance(fr1,fr2,&(ph1->dimensions)) + MINOF(DTW_M[i-1][j  ],
									     DTW_M[i  ][j-1],
									     DTW_M[i-1][j-1]);
  dtw_distance = DTW_M[ph1->frameAmount][ph2->frameAmount];

  backtrackPath2mat(m_cell, DTW_M, ph1->frameAmount, ph2->frameAmount);

  for(i=0; i <= ph1->frameAmount; i++)
    free(DTW_M[i]);
  free(DTW_M);

  return dtw_distance;
}

/* for backtracking the used DTW-path */
void backtrackPath(phoneme *ph, double **DTW_M, int n, int m)
{
  int i,j,k;

  i = n; j = m; k = 0;
  while ( (i > 0) || (j > 0) )
    {
      if ((DTW_M[i-1][j-1] <= DTW_M[i][j-1]) && (DTW_M[i-1][j-1] <= DTW_M[i-1][j]))
	{
	  ph->DTWpath[k++] = '3';
	  i--;
	  j--;
	}
      else if (DTW_M[i][j-1] <= DTW_M[i-1][j])
	{
	  ph->DTWpath[k++] = '2';
	  j--;
	}
      else 
	{
	  ph->DTWpath[k++] = '1';
	  i--;
	}
    }

  while (ph->DTWpath[k] != '\0')
    ph->DTWpath[k++] = '\0';
}


/* for backtracking the used DTW-path to path-matrix, 
   used in storing the paths for one cluster */
void backtrackPath2mat(char *cell, double **DTW_M, int n, int m)
{
  int i,j,k;

  i = n; j = m; k = 0;
  while ( (i > 0) || (j > 0) )
    {
      if ((DTW_M[i-1][j-1] <= DTW_M[i][j-1]) && (DTW_M[i-1][j-1] <= DTW_M[i-1][j]) )
	{
	  cell[k++] = '3';
	  i--;
	  j--;
	}
      else if (DTW_M[i][j-1] <= DTW_M[i-1][j])
	{
	  cell[k++] = '2';
	  j--;
	}
      else 
	{
	  cell[k++] = '1';
	  i--;
	}
    }

  while (cell[k] != '\0')
    cell[k++] = '\0';
}

/* for getting distance matrix of given set of phonemes, used to define cluster distortion */
double** getDistanceMatrix(phoneme **phs, unsigned int n) {

  unsigned int i,j;
  double **matrix = (double**)malloc(n * sizeof(double*));

  for ( i = 0; i < n; i++)
    matrix[i] = (double*)calloc(n,sizeof(double));

  for ( i = 0; i < n; i++)
    {
      for ( j = i+1; j < n; j++)
        {
	  matrix[i][j] = DTW(phs[i], phs[j]);
	  matrix[j][i] = matrix[i][j];
        }
    }

  return matrix;
}

/* for getting distance vector of given set of phonemes, 
   vector consists of the distances to best medoid-ts within cluster */
double* getDistanceVector(phoneme **phs, unsigned int n, int *idx) 
{
  int i,j;
  double best_sum,sum;
  double *vector = (double*)calloc(n,sizeof(double));
  double *best = (double*)calloc(n,sizeof(double));

  best_sum = DBL_MAX;

  for ( i = 0, sum = 0.0; i < n; i++, sum = 0.0)
    {
      for ( j = 0; j < n; j++)
        {
	  vector[j] = DTW(phs[i], phs[j]);
	  sum += vector[j];
	}
      
      if (best_sum > sum)
	{
	  for ( j = 0; j < n; j++)
	    best[j] = vector[j];
	  best_sum = sum;
	  *idx = i;
	}
    }

  free(vector);

  return best;
}

/* for getting distance vector of given set of phonemes, 
   vector consists of the distances to best medoid-ts within cluster */
double* getDistanceVectorWithPath(phoneme **phs, unsigned int n, int *idx, char **paths, int plen) 
{
  int i,j;
  double best_sum,sum;
  char **tmp = (char**)malloc(n*sizeof(char*));
  double *vector = (double*)calloc(n,sizeof(double));
  double *best = (double*)calloc(n,sizeof(double));

  best_sum = DBL_MAX;

  for(i=0; i < n; i++)
    tmp[i] = (char*)calloc(plen,sizeof(char));

  for ( i = 0, sum = 0.0; i < n; i++, sum = 0.0)
    {
      for ( j = 0; j < n; j++)
        {
	  /*	  if (i == 0)
		  printf("%d\n",j); */
	  vector[j] = DTWpath2mat(phs[j], phs[i], tmp[j]);
	  sum += vector[j];
	}
      
      if (best_sum > sum)
	{
	  for ( j = 0; j < n; j++)
	    {
	      best[j] = vector[j];
	      copyPaths(paths[j],tmp[j]);
	    }
	  best_sum = sum;
	  *idx = i;
	}
    }

  for(i=0; i < n; i++)
    free(tmp[i]);
  free(tmp);
  free(vector);

  return best;
}

/* for getting distance matrix of given set of phonemes, used to define cluster distortion */
double** getDistanceMatrixWithPath(phoneme **phs, unsigned int n, char ***p) {

  unsigned int i,j;
  double **matrix = (double**)malloc(n * sizeof(double*));

  for ( i = 0; i < n; i++)
    matrix[i] = (double*)calloc(n,sizeof(double));

  for ( i = 0; i < n; i++)
    {
      for ( j = i; j < n; j++)
        {
	  matrix[i][j] = DTWpath2mat(phs[i], phs[j], p[i][j]);
	  if (i != j)
	    matrix[j][i] = DTWpath2mat(phs[j], phs[i], p[j][i]);
        }
    }

  return matrix;
}

/* for calculating cluster's internal distortion with 
   help of table created while assigning medoids */
void calculateIntDist(cluster *cs, double value)
{
  cs->distortion = value / ((double)(cs->phAmount));
}

/* for calculating cluster's internal distortion from scratch */
void calculateInternalDistortion(cluster *cs)
{
  int i;
  double distortion = 0.0;
  for (i=0; i < cs->phAmount; i++)
    {
      cs->phs[i]->distortion = DTW(cs->exemplar,cs->phs[i]); 
      distortion += cs->phs[i]->distortion;
    }

  cs->distortion =  distortion / ((double)(cs->phAmount));
}

double calculateExternalDistortion(int n, cluster **cs, int m)
{
  int i,j;
  double distortion = 0.0;
  for (i=0; i < m; i++)
    for (j=0; j < cs[i]->phAmount; j++) 
	distortion += cs[i]->phs[j]->distortion;

  return distortion / ((double)n);
}

/* for calculating internal & external distortions */
double calculateDistortion(int n, cluster **cs, int m)
{
  int i;
  for (i=0; i < m; i++)
    calculateInternalDistortion(cs[i]);

  return calculateExternalDistortion(n,cs,m);
}

double ClassificationError(phoneme **phs, phoneme **ts, int n)
{
  int i, j;
  int error = 0;
  int total = 0;
  int correct;
  int clustering;

  for (i=0; i < n; i++)
    {
      for (j=i+1; j < n; j++)
	{
	  correct = (ts[i]->myCluster == ts[j]->myCluster);
	  clustering = (phs[i]->myCluster == phs[j]->myCluster);

	  if (correct != clustering)
	    {
	      error++;
	    }

	  total++;
	}
    }

  return (double)error / (double)total;
}
